@@ -134,7 +134,7 @@ import Foreign.StablePtr
 #if (__GLASGOW_HASKELL__ >= 707) || (defined(MIN_VERSION_template_haskell) && USE_TYPE_LITS)
 import GHC.TypeLits
 # if MIN_VERSION_base(4,10,0)
-import Numeric.Natural (Natural)
+import qualified Numeric.Natural as Numeric (Natural)
 # elif __GLASGOW_HASKELL__ >= 707
 import Control.Exception (ArithException(..), throw)
 # endif
@@ -154,8 +154,12 @@ import System.IO.Unsafe
 import Unsafe.Coerce
 #endif
 
-#ifdef HLINT
-{-# ANN module "HLint: ignore Avoid lambda" #-}
+#if MIN_VERSION_base(4,7,0)
+import Data.Coerce (Coercible, coerce)
+#endif
+
+#if MIN_VERSION_base(4,18,0)
+import qualified GHC.TypeNats as TN
 #endif
 
 -- Due to https://gitlab.haskell.org/ghc/ghc/issues/16893, inlining
@@ -199,8 +203,6 @@ instance KnownSymbol n => Reifies n String where
 -- KnownNat
 --------------------------------------------------------------------------------
 
-newtype MagicNat r = MagicNat (forall (n :: Nat). KnownNat n => Proxy n -> r)
-
 -- | This upgraded version of 'reify' can be used to generate a 'KnownNat' suitable for use with other APIs.
 --
 -- Attemping to pass a negative 'Integer' as an argument will result in an
@@ -208,6 +210,8 @@ newtype MagicNat r = MagicNat (forall (n :: Nat). KnownNat n => Proxy n -> r)
 --
 -- /Available only on GHC 7.8+/
 --
+-- >>> import GHC.TypeLits
+--
 -- >>> reifyNat 4 natVal
 -- 4
 --
@@ -215,37 +219,58 @@ newtype MagicNat r = MagicNat (forall (n :: Nat). KnownNat n => Proxy n -> r)
 -- 4
 
 reifyNat :: forall r. Integer -> (forall (n :: Nat). KnownNat n => Proxy n -> r) -> r
+# if MIN_VERSION_base(4,18,0)
+-- With base-4.18 or later, we can use the API in GHC.TypeNats to define this
+-- function directly.
+reifyNat n k = TN.withSomeSNat (fromInteger n :: Numeric.Natural) $
+               \(sn :: (SNat n)) -> TN.withKnownNat sn $ k (Proxy :: Proxy n)
+{-# INLINE reifyNat #-}
+# else
+-- On older versions of base, we resort to unsafeCoerce.
 reifyNat n k = unsafeCoerce (MagicNat k :: MagicNat r)
-# if MIN_VERSION_base(4,10,0)
+#  if MIN_VERSION_base(4,10,0)
                              -- Starting with base-4.10, the internal
                              -- representation of KnownNat changed from Integer
                              -- to Natural, so make sure to perform the same
                              -- conversion before unsafeCoercing.
-                             (fromInteger n :: Natural)
-# else
+                             (fromInteger n :: Numeric.Natural)
+#  else
                              (if n < 0 then throw Underflow else n)
-# endif
+#  endif
                              Proxy
 {-# INLINE_UNSAFE_COERCE reifyNat #-}
 
+newtype MagicNat r = MagicNat (forall (n :: Nat). KnownNat n => Proxy n -> r)
+# endif
+
 --------------------------------------------------------------------------------
 -- KnownSymbol
 --------------------------------------------------------------------------------
 
-newtype MagicSymbol r = MagicSymbol (forall (n :: Symbol). KnownSymbol n => Proxy n -> r)
-
 -- | This upgraded version of 'reify' can be used to generate a 'KnownSymbol' suitable for use with other APIs.
 --
 -- /Available only on GHC 7.8+/
 --
+-- >>> import GHC.TypeLits
+--
 -- >>> reifySymbol "hello" symbolVal
 -- "hello"
 --
 -- >>> reifySymbol "hello" reflect
 -- "hello"
 reifySymbol :: forall r. String -> (forall (n :: Symbol). KnownSymbol n => Proxy n -> r) -> r
+# if MIN_VERSION_base(4,18,0)
+-- With base-4.18 or later, we can use the API in GHC.TypeNats to define this
+-- function directly.
+reifySymbol s k = withSomeSSymbol s $ \(ss :: SSymbol s) -> withKnownSymbol ss (k (Proxy :: Proxy s))
+{-# INLINE reifySymbol #-}
+# else
+-- On older versions of base, we resort to unsafeCoerce.
 reifySymbol n k = unsafeCoerce (MagicSymbol k :: MagicSymbol r) n Proxy
 {-# INLINE_UNSAFE_COERCE reifySymbol #-}
+# endif
+
+newtype MagicSymbol r = MagicSymbol (forall (n :: Symbol). KnownSymbol n => Proxy n -> r)
 #endif
 
 ------------------------------------------------------------------------------
@@ -664,5 +689,10 @@ traverseBy pur app f = reifyApplicative pur app (traverse (ReflectedApplicative
 sequenceBy :: Traversable t => (forall x. x -> f x) -> (forall x y. f (x -> y) -> f x -> f y) -> t (f a) -> f (t a)
 sequenceBy pur app = reifyApplicative pur app (traverse ReflectedApplicative)
 
+#if MIN_VERSION_base(4,7,0)
+(#.) :: Coercible c b => (b -> c) -> (a -> b) -> (a -> c)
+(#.) _ = coerce (\x -> x :: b) :: forall a b. Coercible b a => a -> b
+#else
 (#.) :: (b -> c) -> (a -> b) -> a -> c
 (#.) _ = unsafeCoerce
+#endif
